__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"


def set_pea_parser(parser=None):
    """Set the parser for the Pea

    :param parser: an optional existing parser to build upon
    :return: the parser
    """
    if not parser:
        from .base import set_base_parser

        parser = set_base_parser()

    from .peapods.base import mixin_base_ppr_parser
    from .peapods.runtimes.zmq import mixin_zmq_runtime_parser
    from .peapods.runtimes.zed import mixin_zed_runtime_parser
    from .peapods.runtimes.container import mixin_container_runtime_parser
    from .peapods.runtimes.remote import mixin_remote_parser
    from .peapods.pea import mixin_pea_parser
    from .peapods.runtimes.distributed import mixin_distributed_feature_parser

    mixin_base_ppr_parser(parser)
    mixin_zmq_runtime_parser(parser)
    mixin_zed_runtime_parser(parser)
    mixin_container_runtime_parser(parser)
    mixin_remote_parser(parser)
    mixin_distributed_feature_parser(parser)
    mixin_pea_parser(parser)

    return parser


def set_pod_parser(parser=None):
    """Set the parser for the Pod

    :param parser: an optional existing parser to build upon
    :return: the parser
    """
    if not parser:
        from .base import set_base_parser

        parser = set_base_parser()

    set_pea_parser(parser)

    from .peapods.pod import mixin_base_pod_parser

    mixin_base_pod_parser(parser)

    return parser


def set_gateway_parser(parser=None):
    """Set the parser for the gateway arguments

    :param parser: an optional existing parser to build upon
    :return: the parser
    """
    if not parser:
        from .base import set_base_parser

        parser = set_base_parser()

    from .peapods.base import mixin_base_ppr_parser
    from .peapods.runtimes.zmq import mixin_zmq_runtime_parser
    from .peapods.runtimes.zed import mixin_zed_runtime_parser
    from .peapods.runtimes.container import mixin_container_runtime_parser
    from .peapods.runtimes.remote import mixin_remote_parser
    from .peapods.runtimes.remote import mixin_grpc_parser
    from .peapods.pea import mixin_pea_parser

    mixin_base_ppr_parser(parser)
    mixin_zmq_runtime_parser(parser)
    mixin_zed_runtime_parser(parser)
    mixin_grpc_parser(parser)
    mixin_remote_parser(parser)
    mixin_pea_parser(parser)

    from ..enums import SocketType, PodRoleType

    parser.set_defaults(
        name='gateway',
        socket_in=SocketType.PULL_CONNECT,  # otherwise there can be only one client at a time
        socket_out=SocketType.PUSH_CONNECT,
        ctrl_with_ipc=True,  # otherwise ctrl port would be conflicted
        read_only=True,
        runtime_cls='GRPCRuntime',
        pod_role=PodRoleType.GATEWAY,
    )
    return parser


def set_client_cli_parser(parser=None):
    """Set the parser for the cli client

    :param parser: an optional existing parser to build upon
    :return: the parser
    """
    if not parser:
        from .base import set_base_parser

        parser = set_base_parser()

    from .peapods.runtimes.remote import mixin_grpc_parser, mixin_remote_parser
    from .client import mixin_client_cli_parser

    mixin_client_cli_parser(parser)
    mixin_grpc_parser(parser)
    mixin_remote_parser(parser)

    return parser


def get_main_parser():
    """The main parser for Jina

    :return: the parser
    """
    from .base import set_base_parser
    from .helloworld import set_hw_parser, set_hello_parser
    from .helper import _chf, _SHOW_ALL_ARGS
    from .check import set_check_parser
    from .export_api import set_export_api_parser
    from .flow import set_flow_parser
    from .hub import set_hub_parser
    from .ping import set_ping_parser
    from .optimizer import set_optimizer_parser

    # create the top-level parser
    parser = set_base_parser()

    sp = parser.add_subparsers(
        dest='cli',
        description='use `%(prog)-8s [sub-command] --help` '
        'to get detailed information about each sub-command',
        required=True,
    )

    set_hello_parser(
        sp.add_parser(
            'hello',
            help='👋 Hello World! Hello Jina!',
            description='Start hello-world demos',
            formatter_class=_chf,
        )
    )

    set_pod_parser(
        sp.add_parser(
            'pod',
            help='Start a Pod',
            description='Start a Jina Pod',
            formatter_class=_chf,
        )
    )

    set_flow_parser(
        sp.add_parser(
            'flow',
            description='Start a Flow that orchestrates multiple pods',
            help='Start a Flow',
            formatter_class=_chf,
        )
    )

    set_optimizer_parser(
        sp.add_parser(
            'optimizer',
            description='Start a FlowOptimizer from a YAML configuration file',
            help='Start an FlowOptimizer from a YAML file',
            formatter_class=_chf,
        )
    )

    set_gateway_parser(
        sp.add_parser(
            'gateway',
            description='Start a Gateway that receives client Requests via gRPC/REST interface',
            help='Start a Gateway',
            formatter_class=_chf,
        )
    )

    set_ping_parser(
        sp.add_parser(
            'ping',
            help='Ping a pod and check its connectivity',
            description='Ping a remote pod and check the network connectivity',
            formatter_class=_chf,
        )
    )

    set_check_parser(
        sp.add_parser(
            'check',
            help='Check the import of all Executors and Drivers',
            description='Check the import status of all executors and drivers',
            formatter_class=_chf,
        )
    )

    set_hub_parser(
        sp.add_parser(
            'hub',
            help='Build, push, pull Jina Hub images',
            description='Build, push, pull Jina Hub images',
            formatter_class=_chf,
        )
    )

    # Below are low-level / internal / experimental CLIs, hidden from users by default

    set_pea_parser(
        sp.add_parser(
            'pea',
            description='Start a Jina pea. '
            'You should rarely use this directly unless you '
            'are doing low-level orchestration',
            formatter_class=_chf,
            **(dict(help='start a pea')) if _SHOW_ALL_ARGS else {}
        )
    )

    set_client_cli_parser(
        sp.add_parser(
            'client',
            description='Start a Python client that connects to a remote Jina gateway',
            formatter_class=_chf,
            **(dict(help='start a client')) if _SHOW_ALL_ARGS else {}
        )
    )

    set_export_api_parser(
        sp.add_parser(
            'export-api',
            description='Export Jina API to JSON/YAML file for 3rd party applications',
            formatter_class=_chf,
            **(dict(help='export Jina API to file')) if _SHOW_ALL_ARGS else {}
        )
    )

    set_hw_parser(
        sp.add_parser(
            'hello-world',
            description='Start the hello-world demo, a simple end2end image index and search demo '
            'without any extra dependencies.',
            formatter_class=_chf,
            **(dict(help='👋 Hello World! Hello Jina!')) if _SHOW_ALL_ARGS else {}
        )
    )

    return parser



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from typing import Optional, Iterable

from jina.executors.indexers.keyvalue import BinaryPbIndexer

if False:
    from redis import Redis


class RedisDBIndexer(BinaryPbIndexer):
    """
    :class:`RedisDBIndexer` Use Redis as a key-value indexer.
    """

    def __init__(self,
                 hostname: str = '0.0.0.0',
                 # default port on linux
                 port: int = 6379,
                 db: int = 0,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.hostname = hostname
        self.port = port
        self.db = db

    def get_query_handler(self) -> 'Redis':
        """Get the database handler.
        """
        import redis
        try:
            r = redis.Redis(host=self.hostname, port=self.port, db=self.db, socket_timeout=10)
            r.ping()
            return r
        except redis.exceptions.ConnectionError as r_con_error:
            self.logger.error('Redis connection error: ', r_con_error)
            raise

    def query(self, key: str, *args, **kwargs) -> Optional[bytes]:
        """Find the protobuf document via id.
        :param key: ``id``
        :return: matching document
        """
        with self.get_query_handler() as redis_handler:
            return redis_handler.get(key)

    def add(self, keys: Iterable[str], values: Iterable[bytes], *args, **kwargs) -> None:
        """Add JSON-friendly serialized documents to the index.

        :param keys: document ids
        :param values: JSON-friendly serialized documents
        """
        redis_docs = [{'_id': i, 'values': j} for i, j in zip(keys, values)]

        with self.get_query_handler() as redis_handler:
            for k in redis_docs:
                redis_handler.set(k['_id'], k['values'])

    def update(self, keys: Iterable[str], values: Iterable[bytes], *args, **kwargs) -> None:
        """Update JSON-friendly serialized documents on the index.

        :param keys: document ids to update
        :param values: JSON-friendly serialized documents
        """
        missed = []
        for key in keys:
            if self.query(key) is None:
                missed.append(key)
        if missed:
            raise KeyError(f'Key(s) {missed} were not found in redis')

        self.delete(keys)
        self.add(keys, values)

    def delete(self, keys: Iterable[str], *args, **kwargs) -> None:
        """Delete documents from the index.

        :param keys: document ids to delete
        """
        with self.get_query_handler() as h:
            for k in keys:
                h.delete(k)



"""Module wrapping the Client of Jina."""
__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from . import request
from .base import BaseClient, CallbackFnType, InputType, InputDeleteType
from .helper import callback_exec
from .request import GeneratorSourceType
from .websocket import WebSocketClientMixin
from ..enums import RequestType
from ..helper import run_async, deprecated_alias


class Client(BaseClient):
    """A simple Python client for connecting to the gRPC gateway.

    It manages the asyncio event loop internally, so all interfaces are synchronous from the outside.
    """

    async def _get_results(self, *args, **kwargs):
        result = []
        async for resp in super()._get_results(*args, **kwargs):
            if self.args.return_results:
                result.append(resp)

        if self.args.return_results:
            return result

    @deprecated_alias(
        input_fn=('inputs', 0),
        buffer=('inputs', 1),
        callback=('on_done', 1),
        output_fn=('on_done', 1),
    )
    def train(
        self,
        inputs: InputType,
        on_done: CallbackFnType = None,
        on_error: CallbackFnType = None,
        on_always: CallbackFnType = None,
        **kwargs
    ) -> None:
        """Issue 'train' request to the Flow.

        :param inputs: input data which can be an Iterable, a function which returns an Iterable, or a single Document
        :param on_done: the function to be called when the :class:`Request` object is resolved.
        :param on_error: the function to be called when the :class:`Request` object is rejected.
        :param on_always: the function to be called when the :class:`Request` object is  is either resolved or rejected.
        :param kwargs: additional parameters
        :return: None
        """
        self.mode = RequestType.TRAIN
        return run_async(
            self._get_results, inputs, on_done, on_error, on_always, **kwargs
        )

    @deprecated_alias(
        input_fn=('inputs', 0),
        buffer=('inputs', 1),
        callback=('on_done', 1),
        output_fn=('on_done', 1),
    )
    def search(
        self,
        inputs: InputType,
        on_done: CallbackFnType = None,
        on_error: CallbackFnType = None,
        on_always: CallbackFnType = None,
        **kwargs
    ) -> None:
        """Issue 'search' request to the Flow.

        :param inputs: input data which can be an Iterable, a function which returns an Iterable, or a single Document
        :param on_done: the function to be called when the :class:`Request` object is resolved.
        :param on_error: the function to be called when the :class:`Request` object is rejected.
        :param on_always: the function to be called when the :class:`Request` object is  is either resolved or rejected.
        :param kwargs: additional parameters
        :return: None
        """
        self.mode = RequestType.SEARCH
        self.add_default_kwargs(kwargs)
        return run_async(
            self._get_results, inputs, on_done, on_error, on_always, **kwargs
        )

    @deprecated_alias(
        input_fn=('inputs', 0),
        buffer=('inputs', 1),
        callback=('on_done', 1),
        output_fn=('on_done', 1),
    )
    def index(
        self,
        inputs: InputType,
        on_done: CallbackFnType = None,
        on_error: CallbackFnType = None,
        on_always: CallbackFnType = None,
        **kwargs
    ) -> None:
        """Issue 'index' request to the Flow.

        :param inputs: input data which can be an Iterable, a function which returns an Iterable, or a single Document
        :param on_done: the function to be called when the :class:`Request` object is resolved.
        :param on_error: the function to be called when the :class:`Request` object is rejected.
        :param on_always: the function to be called when the :class:`Request` object is  is either resolved or rejected.
        :param kwargs: additional parameters
        :return: None
        """
        self.mode = RequestType.INDEX
        return run_async(
            self._get_results, inputs, on_done, on_error, on_always, **kwargs
        )

    @deprecated_alias(
        input_fn=('inputs', 0),
        buffer=('inputs', 1),
        callback=('on_done', 1),
        output_fn=('on_done', 1),
    )
    def update(
        self,
        inputs: InputType,
        on_done: CallbackFnType = None,
        on_error: CallbackFnType = None,
        on_always: CallbackFnType = None,
        **kwargs
    ) -> None:
        """Issue 'update' request to the Flow.

        :param inputs: input data which can be an Iterable, a function which returns an Iterable, or a single Document
        :param on_done: the function to be called when the :class:`Request` object is resolved.
        :param on_error: the function to be called when the :class:`Request` object is rejected.
        :param on_always: the function to be called when the :class:`Request` object is  is either resolved or rejected.
        :param kwargs: additional parameters
        :return: None
        """
        self.mode = RequestType.UPDATE
        return run_async(
            self._get_results, inputs, on_done, on_error, on_always, **kwargs
        )

    @deprecated_alias(
        input_fn=('inputs', 0),
        buffer=('inputs', 1),
        callback=('on_done', 1),
        output_fn=('on_done', 1),
    )
    def delete(
        self,
        inputs: InputDeleteType,
        on_done: CallbackFnType = None,
        on_error: CallbackFnType = None,
        on_always: CallbackFnType = None,
        **kwargs
    ) -> None:
        """Issue 'update' request to the Flow.

        :param inputs: input data which can be an Iterable, a function which returns an Iterable, or a single Document id.
        :param on_done: the function to be called when the :class:`Request` object is resolved.
        :param on_error: the function to be called when the :class:`Request` object is rejected.
        :param on_always: the function to be called when the :class:`Request` object is  is either resolved or rejected.
        :param kwargs: additional parameters
        :return: None
        """
        self.mode = RequestType.DELETE
        return run_async(
            self._get_results, inputs, on_done, on_error, on_always, **kwargs
        )


class WebSocketClient(Client, WebSocketClientMixin):
    """A Python Client to stream requests from a Flow with a REST Gateway.

    :class:`WebSocketClient` shares the same interface as :class:`Client` and provides methods like
    :meth:`index`, "meth:`search`, :meth:`train`, :meth:`update` & :meth:`delete`.

    It is used by default while running operations when we create a `Flow` with `rest_api=True`

    .. highlight:: python
    .. code-block:: python

        from jina.flow import Flow
        f = Flow(rest_api=True).add().add()

        with f:
            f.index(['abc'])


    :class:`WebSocketClient` can also be used to run operations for a remote Flow

    .. highlight:: python
    .. code-block:: python

        # A Flow running on remote
        from jina.flow import Flow
        f = Flow(rest_api=True, port_expose=34567).add().add()

        with f:
            f.block()

        # Local WebSocketClient running index & search
        from jina.clients import WebSocketClient

        client = WebSocketClient(...)
        client.index(...)
        client.search(...)


    :class:`WebSocketClient` internally handles an event loop to run operations asynchronously.
    """
